package org.artifactory.engine;

import org.artifactory.addon.AddonsManager;
import org.artifactory.aol.AolService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.fs.RepoResource;
import org.artifactory.repo.Repo;
import org.artifactory.request.ArtifactoryRequest;
import org.artifactory.request.InternalRequestContext;
import org.artifactory.resource.RepoResourceInfo;
import org.artifactory.security.access.AccessTokenAuthentication;
import org.artifactory.signature.RedirectType;
import org.artifactory.storage.binstore.service.InternalBinaryService;
import org.jfrog.access.token.JwtAccessToken;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.AssertJUnit.assertEquals;

/**
 * @author Shay Bagants
 */
@Test
public class DownloadServiceImplTest {

    private DownloadServiceImpl downloadService = new DownloadServiceImpl();

    @Mock
    private InternalRequestContext requestContext;
    @Mock
    private ArtifactoryRequest request;
    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private DownloadRedirectHelper downloadRedirectHelper;
    @Mock
    private InternalBinaryService binaryService;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private AolService aolService;

    @BeforeMethod
    private void setup() {
        MockitoAnnotations.initMocks(this);
        when(requestContext.getRequest()).thenReturn(request);

        downloadService.authorizationService = authorizationService;
        downloadService.downloadRedirectHelper = downloadRedirectHelper;
        downloadService.addonsManager = addonsManager;
        downloadService.binaryService = binaryService;

        ArtifactoryContext context = mock(ArtifactoryContext.class);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    private void unbind() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test(dataProvider = "provideRequestData")
    public void testGetXrayUserName(String userAgent, String principal, String expectedResult) {
        when(authorizationService.currentUsername()).thenReturn(principal);
        when(request.getHeader("User-Agent")).thenReturn(userAgent);
        String result = downloadService.getXrayUserName(requestContext, null);
        Assert.assertEquals(result, expectedResult);
    }

    @Test(dataProvider = "provideRequestDataForAccessTokenAuthentication", description = "RTFACT-20177")
    public void testGetXrayUserNameUsingAccessTokenAuthentication(String userAgent, String principal,
            String expectedResult) {
        AccessTokenAuthentication accessTokenAuthentication = mock(AccessTokenAuthentication.class);
        JwtAccessToken jwtAccessToken = mock(JwtAccessToken.class);
        when(accessTokenAuthentication.getAccessToken()).thenReturn(jwtAccessToken);
        when(jwtAccessToken.getSubject()).thenReturn(principal);
        when(request.getHeader("User-Agent")).thenReturn(userAgent);
        String result = downloadService.getXrayUserName(requestContext, accessTokenAuthentication);
        Assert.assertEquals(result, expectedResult);
    }

    @Test(dataProvider = "getCdnConfigurationsData")
    public void testGetRedirectTypeForRepo_differentCdnConfigurations(boolean isCdn, boolean isEplus, boolean isEdge,
            boolean isEnterprise, boolean isAol, boolean isAnonymous, RedirectType expectedType) {
        setupCdn();

        RepoResource resource = mock(RepoResource.class);
        Repo repo = mock(Repo.class);
        LocalRepoDescriptor repoDescriptor = mock(LocalRepoDescriptor.class);
        RepoResourceInfo repoResourceInfo = mock(RepoResourceInfo.class);

        when(repoResourceInfo.getSha1()).thenReturn("cf23df2207d99a74fbe169e3eba035e633b65d94");
        when(resource.getInfo()).thenReturn(repoResourceInfo);
        when(repoDescriptor.isReal()).thenReturn(true);
        when(repo.getDescriptor()).thenReturn(repoDescriptor);

        // using params
        when(repoDescriptor.isCdnRedirect()).thenReturn(isCdn);
        when(addonsManager.isEnterprisePlusInstalled()).thenReturn(isEplus);
        when(addonsManager.isBasicEnterpriseLicensed()).thenReturn(isEnterprise);
        when(addonsManager.isEdgeLicensed()).thenReturn(isEdge);
        when(aolService.isAol()).thenReturn(isAol);
        when(authorizationService.isAnonymous()).thenReturn(isAnonymous);

        RedirectType redirectType = downloadService.getRedirectTypeForRepo(requestContext, repo, resource);

        assertEquals(expectedType, redirectType);
    }

    @Test(dataProvider = "shouldRedirectProvider")
    public void testShouldRedirect(boolean headRequest, boolean expectedResult) {
        LocalRepoDescriptor repoDescriptor = mock(LocalRepoDescriptor.class);
        when(repoDescriptor.isReal()).thenReturn(true);
        when(downloadRedirectHelper.isRedirectEnable(any(), any())).thenReturn(true);
        when(downloadRedirectHelper.isArtifactGreaterEqualsRedirectThreshold(any())).thenReturn(true);
        when(binaryService.canCloudProviderGenerateRedirection(any())).thenReturn(true);
        when(request.isHeadOnly()).thenReturn(headRequest);
        RepoResource resource = mock(RepoResource.class);
        RepoResourceInfo repoResourceInfo = mock(RepoResourceInfo.class);
        when(repoResourceInfo.getSha1()).thenReturn("cf23df2207d99a74fbe169e3eba035e633b65d94");
        when(resource.getInfo()).thenReturn(repoResourceInfo);

        boolean shouldRedirect = downloadService.shouldRedirect(requestContext, null, resource, repoDescriptor, true);
        assertEquals(shouldRedirect, expectedResult);
    }

    @DataProvider
    public Object[][] shouldRedirectProvider() {
        return new Object[][] {
                {true, false},
                {false, true}
        };
    }

    @DataProvider
    public static Object[][] provideRequestData() {
        return new Object[][] {
                //agent, user, expectedResult
                {"xray/1.0", "xray", "xray"},
                {"xray/1.0", "test-user", null},
                {"jfrog-cli", "xray", null},
                {"xray/1.0", "token:jfrt@123", null},
                {"jfrog-cli", "token:jfxr@123", null},
                {"x-ray/1.0", "token:jfxr@123", null},
        };
    }

    @DataProvider
    public static Object[][] provideRequestDataForAccessTokenAuthentication() {
        return new Object[][] {
                //agent, user, expectedResult
                {"xray/1.0", "xray", null},
                {"xray/1.0", "test-user", null},
                {"jfrog-cli", "xray", null},
                {"xray/1.0", "token:jfrt@123", null},
                {"jfrog-cli", "jfxr@123", null},
                {"xray/1.0", "jfxr@123", "jfxr@123"},
                {"x-ray/1.0", "jfxidx@123", null},
                {"xray/1.0", "jfxidx@123", "jfxidx@123"},
                {"x-ray/1.0", "jfxpst@123", null},
                {"xray/1.0", "jfxpst@123", "jfxpst@123"},
                {"x-ray/1.0", "jfxana@123", null},
                {"xray/1.0", "jfxana@123", "jfxana@123"},
        };
    }

    @DataProvider
    public static Object[][] getCdnConfigurationsData() {
        return new Object[][] {
                //isCdn, isEplus, isEdge, isEnterprise, isAol, isAnonymous, expectedType
                //AOL
                {true, true, false, false, true, false, RedirectType.CLOUD_FRONT},
                {true, false, true, false, true, false, RedirectType.CLOUD_FRONT},
                {true, true, false, false, true, true, RedirectType.CLOUD_FRONT},
                {true, false, true, false, true, true, RedirectType.CLOUD_FRONT},
                {true, false, false, true, true, true, RedirectType.CLOUD_FRONT},
                {true, false, false, true, true, false, RedirectType.S3},
                {true, false, false, false, true, true, RedirectType.S3},
                {true, false, false, false, true, false, RedirectType.S3},
                {false, true, false, false, true, false, RedirectType.S3},
                {false, false, true, false, true, false, RedirectType.S3},
                {false, false, false, false, true, false, RedirectType.S3},
                //OnPrem - isCdn cannot be true onPrem (as of 29/03/20)
                {false, true, false, false, false, false, RedirectType.S3},
                {false, false, true, false, false, false, RedirectType.S3},
                {false, false, false, false, false, false, RedirectType.DIRECT},
                {false, false, false, true, false, true, RedirectType.DIRECT},
                {false, false, false, true, false, false, RedirectType.DIRECT}
        };
    }

    private void setupCdn() {
        when(downloadRedirectHelper.isRedirectEnable(any(), any())).thenReturn(true);
        when(downloadRedirectHelper.isArtifactGreaterEqualsRedirectThreshold(any())).thenReturn(true);
        when(binaryService.canCloudProviderGenerateRedirection(any())).thenReturn(true);
        when(aolService.isAol()).thenReturn(true);
    }
}