package org.artifactory.security.access;

import com.google.common.collect.ImmutableList;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.model.xstream.security.GroupImpl;
import org.artifactory.model.xstream.security.UserImpl;
import org.artifactory.security.MutableUserInfo;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.token.JwtAccessToken;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.List;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertTrue;
import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertFalse;

/**
 * @author Shay Bagants
 */
@Test
public class AccessTokenAuthenticationProviderTest {

    private AccessTokenAuthenticationProvider tokenAuthenticationProvider;

    @Mock
    private AccessService accessService;

    @Mock
    private UserGroupService userGroupService;

    @BeforeMethod
    public void setup() {
        MockitoAnnotations.initMocks(this);

        tokenAuthenticationProvider = new AccessTokenAuthenticationProvider(accessService, null, userGroupService);
        ServiceId serviceId = new ServiceId("jfrt", "xyz");
        when(accessService.getArtifactoryServiceId()).thenReturn(serviceId);
    }

    @Test(dataProvider = "provideScopes")
    public void testSetAdminPrivileges(List<String> scopes, boolean expectedAdmin) {
        UserImpl user = new UserImpl();
        JwtAccessToken token = mock(JwtAccessToken.class);
        when(token.getScope()).thenReturn(scopes);
        tokenAuthenticationProvider.setAdminPrivileges(user, new UserImpl(), token);
        assertEquals(user.isAdmin(), expectedAdmin);
    }

    public void testSetAdminPrivilegesFromOriginalUser() {
        UserImpl user = new UserImpl();
        UserImpl userFromStorage = new UserImpl();
        userFromStorage.setAdmin(true);
        JwtAccessToken token = mock(JwtAccessToken.class);
        when(token.getScope()).thenReturn(ImmutableList.of("applied-permissions/user"));
        tokenAuthenticationProvider.setAdminPrivileges(user, userFromStorage, token);
        assertTrue(user.isAdmin());

        user = new UserImpl();
        when(token.getScope()).thenReturn(ImmutableList.of("applied-permissions/groups:group1"));
        tokenAuthenticationProvider.setAdminPrivileges(user, userFromStorage, token);
        // not an identity token, no admin rights
        assertFalse(user.isAdmin());

        user = new UserImpl();
        when(token.getScope())
                .thenReturn(ImmutableList.of("applied-permissions/user", "applied-permissions/groups:group1"));
        tokenAuthenticationProvider.setAdminPrivileges(user, userFromStorage, token);
        // both identity and non-identity scopes are combined
        assertTrue(user.isAdmin());
    }

    @Test(enabled = false, description = "Flaky. Need to investigate")
    public void testGroupsCombinedFromIdentityAndNonIdentityScopes() {
        UserImpl user = new UserImpl();
        user.setUsername("joe");
        UserImpl userFromStorage = new UserImpl();
        userFromStorage.setUsername("joe");
        JwtAccessToken token = mock(JwtAccessToken.class);

        when(token.getScope())
                .thenReturn(ImmutableList.of("applied-permissions/user", "applied-permissions/groups:group1"));
        when(userGroupService.getAllGroups())
                .thenReturn(ImmutableList.of(new GroupImpl("group1"), new GroupImpl("notARealGroup")));

        MutableUserInfo mutatedUser = tokenAuthenticationProvider.populateGroups(user, userFromStorage, token);

        assertTrue(mutatedUser.getGroups().stream().anyMatch(group -> "group1".equals(group.getGroupName())));
        assertTrue(mutatedUser.getGroups().stream().noneMatch(group -> "notARealGroup".equals(group.getGroupName())));
    }

    @DataProvider
    public static Object[][] provideScopes() {
        return new Object[][] {
                {ImmutableList.of("applied-permissions/admin"), true},
                {ImmutableList.of("api:*", "applied-permissions/admin"), true},
                {ImmutableList.of("applied-permissions/nonadmin"), false},
                {ImmutableList.of("applied-permissions/adminn"), false},
                {ImmutableList.of("applied-permissions:admin"), false},
                {ImmutableList.of("applied-permission/admin"), false},
                {ImmutableList.of("applied-permission"), false},
                {ImmutableList.of("admin"), false},
                {ImmutableList.of("applied-permissions/user"), false},
                {ImmutableList.of("jfrt@xyz:admin"), true},
                {ImmutableList.of("api:*", "jfrt@xyz:admin"), true},
                {ImmutableList.of("jfxr@xyz:admin"), false},
                {ImmutableList.of("api:*", "jfxr@xyz:admin"), false},
                {ImmutableList.of("jfrt@abcd:admin"), false},
                {ImmutableList.of("api:*", "jfrt@abcd:admin"), false},
                {ImmutableList.of("jfrt@xyz:user"), false},
        };
    }
}