/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.schedule;

import org.jfrog.common.concurrency.ConflictsGuard;
import org.jfrog.common.concurrency.LockingUtilImpl;
import org.quartz.DisallowConcurrentExecution;
import org.quartz.Job;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.quartz.PersistJobDataAfterExecution;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.security.core.Authentication;

import java.util.List;

/**
 * @author yoavl
 */
@PersistJobDataAfterExecution
@DisallowConcurrentExecution
public abstract class QuartzCommand extends TaskCallback<JobExecutionContext> implements Job {
    private static final Logger log = LoggerFactory.getLogger(QuartzCommand.class);

    @Override
    public final void execute(final JobExecutionContext jobContext) throws JobExecutionException {
        boolean shouldExecute = beforeExecute(jobContext);
        String token = currentTaskToken();
        String taskName = getClass().getName();
        try {
            if (!shouldExecute) {
                log.debug("Skipping execution of task {} for token {}.", taskName, token);
            } else {
                String lockKey = getCommandLockKey();
                if (isClusterSingleton(jobContext)) {
                    doExecuteInLock(jobContext, token, taskName, lockKey, getTaskService().getConflictsGuard());
                } else {
                    doExecute(jobContext, token, lockKey);
                }
            }
        } finally {
            afterExecute();
        }
    }

    public String getCommandLockKey() {
        return ((InternalTask) getActiveTask()).getTaskLockKeyIncludingKeyAttributes(getLockKeyAttributes());
    }

    @Override
    protected boolean isRunOnlyOnMaster(JobExecutionContext jobContext) {
        return jobContext.getMergedJobDataMap().getBoolean(Task.TASK_RUN_CLUSTER_SINGLETON);
    }

    private void doExecute(JobExecutionContext jobContext, String token, String taskName) throws JobExecutionException {
        log.debug("Executing task {} with token {}.", taskName, token);
        onExecute(jobContext);
        log.debug("Finished execution of task {} with token {}.", taskName, token);
    }

    private void doExecuteInLock(JobExecutionContext jobContext, String token, String taskName, String lockKey,
            ConflictsGuard<String> conflictsGuard) {
        log.debug("Trying to execute task {} with lock key {} for token {}.", taskName, lockKey, token);
        new LockingUtilImpl().clusterExclusive(conflictsGuard, lockKey, () -> {
            try {
                log.debug("Acquired lock on task {} with lock key {} for token {}.", taskName, lockKey, token);
                onExecute(jobContext);
            } catch (JobExecutionException e) {
                log.error("Unable to execute job", e);
                throw new IllegalStateException("Unable to execute job", e);
            }
        });
        log.debug("Finished execution of task {} with lock key {} for token {}.", taskName, lockKey, token);
    }

    @Override
    protected String triggeringTaskTokenFromWorkContext(JobExecutionContext jobContext) {
        return jobContext.getMergedJobDataMap().getString(Task.TASK_TOKEN);
    }

    @Override
    protected Authentication getAuthenticationFromWorkContext(JobExecutionContext jobContext) {
        return (Authentication) jobContext.getMergedJobDataMap().get(Task.TASK_AUTHENTICATION);
    }

    @Override
    protected boolean isClusterSingleton(JobExecutionContext jobContext) {
        return jobContext.getMergedJobDataMap().getBoolean(Task.TASK_RUN_CLUSTER_SINGLETON);
    }

    @Override
    protected JobName getJobName(JobExecutionContext jobContext) {
        return (JobName) jobContext.getMergedJobDataMap().get(Task.JOB_NAME);
    }

    @Override
    protected boolean isHardAffinity(JobExecutionContext jobExecutionContext) {
        return (boolean) jobExecutionContext.getMergedJobDataMap().get(Task.HARD_AFFINITY);
    }

    @Override
    protected boolean isAllowedAffinity(JobExecutionContext jobExecutionContext) {
        return (boolean) jobExecutionContext.getMergedJobDataMap().get(Task.ALLOWED_AFFINITY);
    }

    /**
     * @return a list of strings that will be incorporated into the cluster-singleton lock of this job
     */
    protected List<String> getLockKeyAttributes() {
        return List.of();
    }
}