package org.artifactory.ui.rest.service.admin.configuration.repositories.util;

import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.projects.ProjectsService;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.ui.rest.model.admin.configuration.repository.info.LocalRepositoryInfo;
import org.artifactory.ui.rest.model.admin.configuration.repository.info.RepositoryInfo;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.List;
import java.util.Set;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;

/**
 * @author BarakH
 */
@Test
public class GetRepositoryInfoServiceTest {
    private enum UserType {SIMPLE, PROJECT_ADMIN, ADMIN}

    private static final String PROJECT_A = "my-project";
    private static final String REPO_A = "repo-a";
    private static final String DEFAULT_PROJECT_REPO = "DEFAULT_PROJECT_REPO";

    @Mock
    private RepositoryService repositoryService;
    @Mock
    private CentralConfigService configService;
    @Mock
    private ProjectsService projectsService;
    @Mock
    private AuthorizationService authorizationService;

    GetRepositoryInfoService getRepositoryInfoService;

    @BeforeMethod
    public void setUp() {
        MockitoAnnotations.initMocks(this);
        getRepositoryInfoService =
                new GetRepositoryInfoService(repositoryService, configService, projectsService, authorizationService);
        LocalRepoDescriptor localRepo = new LocalRepoDescriptor();
        localRepo.setKey("A");
        when(repositoryService.getLocalRepoDescriptors()).thenReturn(List.of(localRepo));
        CentralConfigDescriptor descriptor = mock(CentralConfigDescriptor.class);
        when(descriptor.getMultiLocalReplications("A")).thenReturn(List.of());
        when(configService.getDescriptor()).thenReturn(descriptor);
    }

    @Test(dataProvider = "provideUserType")
    public void testGetRepositoriesInfo(UserType userType, String project, boolean shouldBeExist) {
        ArtifactoryRestRequest restRequest = mock(ArtifactoryRestRequest.class);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        when(restRequest.getPathParamByKey("repoType")).thenReturn(RepoDetailsType.LOCAL_REPO);
        when(restRequest.getPathParamByKey("project")).thenReturn(project);
        switch (userType) {
            case SIMPLE:
                when(authorizationService.isAdmin()).thenReturn(false);
                break;
            case PROJECT_ADMIN:
                when(authorizationService.isAdmin()).thenReturn(false);
                when(projectsService.getProjectRepositories(PROJECT_A, true)).thenReturn(Set.of("A", "C"));
                break;
            case ADMIN:
                when(authorizationService.isAdmin()).thenReturn(true);
                when(projectsService.getProjectRepositories(PROJECT_A, true)).thenReturn(Set.of("A", "C"));
                break;
        }
        getRepositoryInfoService.execute(restRequest, response);
        List<RepositoryInfo> result = (List<RepositoryInfo>) response.getiModelList();
        if (shouldBeExist) {
            assertEquals("A", result.get(0).getRepoKey());
        } else {
            assertTrue(result.isEmpty());
        }
    }

    @Test
    public void testExecuteRepoInfoIsPopulatedWithProjectInfo() {
        ArtifactoryRestRequest restRequest = mock(ArtifactoryRestRequest.class);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        when(restRequest.getPathParamByKey("repoType")).thenReturn(RepoDetailsType.LOCAL_REPO);
        when(restRequest.getPathParamByKey("project")).thenReturn(PROJECT_A);
        when(projectsService.getProjectRepositories(PROJECT_A, true)).thenReturn(Set.of(REPO_A));

        LocalRepoDescriptor repoA = new LocalRepoDescriptor();
        repoA.setType(RepoType.Maven);
        repoA.setKey(REPO_A);
        when(repositoryService.getLocalRepoDescriptors()).thenReturn(List.of(repoA));
        when(projectsService.getProjectOfRepository(REPO_A)).thenReturn(PROJECT_A);
        getRepositoryInfoService.execute(restRequest, response);

        assertEquals(response.getiModelList().size(), 1);
        LocalRepositoryInfo repoInfo = (LocalRepositoryInfo) response.getiModelList().stream().findFirst().orElse(null);
        assertEquals(repoInfo.getProject(), PROJECT_A);
        assertNull(repoInfo.getEnvironments());

        verify(projectsService).getProjectRepositories(PROJECT_A, true);
        verifyNoMoreInteractions(projectsService);
    }

    @Test
    public void testExecuteRepoInfoIsPopulatedDefaultProjectInfo() {
        ArtifactoryRestRequest restRequest = mock(ArtifactoryRestRequest.class);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        when(restRequest.getPathParamByKey("repoType")).thenReturn(RepoDetailsType.LOCAL_REPO);
        when(restRequest.getPathParamByKey("project")).thenReturn(null);
        when(projectsService.getProjectRepositories(ProjectsService.DEFAULT_PROJECT, true))
                .thenReturn(Set.of(DEFAULT_PROJECT_REPO));

        LocalRepoDescriptor repo = new LocalRepoDescriptor();
        repo.setType(RepoType.Maven);
        repo.setKey(DEFAULT_PROJECT_REPO);
        when(repositoryService.getLocalRepoDescriptors()).thenReturn(List.of(repo));
        when(projectsService.getProjectOfRepository(DEFAULT_PROJECT_REPO)).thenReturn(ProjectsService.DEFAULT_PROJECT);
        getRepositoryInfoService.execute(restRequest, response);

        assertEquals(response.getiModelList().size(), 1);
        LocalRepositoryInfo repoInfo = (LocalRepositoryInfo) response.getiModelList().stream().findFirst().orElse(null);
        assertEquals(repoInfo.getProject(), ProjectsService.DEFAULT_PROJECT);
        assertNull(repoInfo.getEnvironments());

        verify(projectsService).getProjectRepositories(ProjectsService.DEFAULT_PROJECT, true);
        verify(projectsService).getProjectOfRepository(DEFAULT_PROJECT_REPO);
        verifyNoMoreInteractions(projectsService);
    }

    @DataProvider
    public static Object[][] provideUserType() {
        return new Object[][] {
                {UserType.SIMPLE, "", false},
                {UserType.SIMPLE, PROJECT_A, false},
                {UserType.PROJECT_ADMIN, "", false},
                {UserType.PROJECT_ADMIN, PROJECT_A, true},
                {UserType.ADMIN, "", true},
                {UserType.ADMIN, PROJECT_A, true},
        };
    }
}